#replace_package_name_marker
name="test-pkg-hmoazam"
#replace_package_version_marker
__version__="5.11.14"
__description__="DSPy"
__url__="https://github.com/stanfordnlp/dspy"
__author__="Omar Khattab"
__author_email__="okhattab@stanford.edu"